import numpy as np
import arcpy, sys, copy
from arcpy.sa import Contour

"""
This tool will have the following funcitonality:

1) A user will be able to supply a raster.
2) The user will be able to specify one or more PVCs to create
3) The user will be able to output contour lines or polygons.

Note, for really big rasters, I have a clever 64-bit script that uses
compressed masked arrays to manage the computation. Email me if required.

"""


# Get the input raster  
in_raster = arcpy.GetParameterAsText(0)

# Get the PVCs to create
pvc = arcpy.GetParameterAsText(1)

# Get the output location (directory)
out_workspace = arcpy.GetParameterAsText(2)

# Create Polylines?
polylines = arcpy.GetParameterAsText(3)
if polylines == "true":
    polylines = True
else:
    polylines = False
# Create Polylines?
polygons = arcpy.GetParameterAsText(4)
if polygons == "true":
    polygons = True
else:
    polygons = False

# Remove holes?
holes = arcpy.GetParameterAsText(5)

if holes == "true":
    holes = True
else:
    holes = False


# Here is a hole removing function - it probably doesn't cover all the cases, caveat emptor. 
def removeholes(poly,line):
    arcpy.env.overwriteOutput = True
    arcpy.MakeFeatureLayer_management(poly,"hole")
    # get a point centroid for each polygon that is definitely inside the polygon.
    temp_out = out_workspace + "\\temp_centroids.shp"
    arcpy.FeatureToPoint_management("hole",temp_out,"INSIDE")
    arcpy.MakeFeatureLayer_management(temp_out,"points")
    # Basically, we want to draw a line from each point, and count the number of times it cuts a boundary.
    # Even numbers should indicate that a point is from a hole polygon.
    
    # Get the point locations, as arcpy point types, and store as a list
    points_list = []
    desc = arcpy.Describe("points")
    shapefieldname = desc.ShapeFieldName
    for row in arcpy.SearchCursor("points"):
        feat = row.getValue(shapefieldname)
        pnt = feat.getPart()
        points_list.append(arcpy.Point(pnt.X,pnt.Y))
    arcpy.Delete_management(temp_out)
    # Work out the max northing of the polygon file (and hence the research domain)
    Yextend = arcpy.Describe("hole").extent.YMax + (0.5 * (arcpy.Describe("hole").extent.YMax-arcpy.Describe("hole").extent.YMin))
    lines_list = []
    for i in xrange(0,len(points_list)):
        start = points_list[i]
        end = arcpy.Point(start.X+1.0,Yextend)
        array = arcpy.Array()
        array.add(start)
        array.add(end)
        lines_list.append(arcpy.Polyline(array))

    arcpy.CopyFeatures_management(lines_list, out_workspace + "\\polyline_File.shp")
    
    arcpy.MakeFeatureLayer_management(line,"lines")

    intersect_counts = []
    
    for i in xrange(0,len(lines_list)):
        arcpy.MakeFeatureLayer_management(out_workspace + "\\polyline_File.shp","test","\"FID\" =" + str(i))
        #lineout = out_workspace + "\\temp_line_" + str(i) + ".shp"
        #arcpy.CopyFeatures_management("test",lineout)
        arcpy.Intersect_analysis(["test","lines"],"temp","ALL",'',"POINT")
        temp_sing = out_workspace + "\\temp_sing.shp"
        arcpy.MultipartToSinglepart_management("temp",temp_sing)
        arcpy.MakeFeatureLayer_management(temp_sing,"temp2")
        intersect_counts.append(int(arcpy.GetCount_management("temp2").getOutput(0)))
    arcpy.Delete_management(out_workspace + "\\polyline_File.shp")
    arcpy.Delete_management(temp_sing)
    # Holes have an even number of intersections, whilst non-holes have odd numbers!

    OIDs_for_holes = []
    
    for i in xrange(0,len(intersect_counts)):
        if (intersect_counts[i] % 2 == 0):
            OIDs_for_holes.append(i)
    #arcpy.AddMessage(str(intersect_counts))
    #arcpy.AddMessage(str(OIDs_for_holes))
    if len(OIDs_for_holes) == 0:
        arcpy.env.overwriteOutput = False
        return
    else:
        arcpy.SelectLayerByAttribute_management("hole","NEW_SELECTION","\"FID\" in (" + str(OIDs_for_holes)[1:-1] + ")")
        #arcpy.AddMessage("\"FID\" in (" + str(OIDs_for_holes)[1:-1] + ")")
        #arcpy.AddMessage(str(int(arcpy.GetCount_management("temp2").getOutput(0))))
        arcpy.DeleteFeatures_management("hole")

        arcpy.env.overwriteOutput = False
        return


# First deal with the Raster, save the details of the Lower Left corner, and cell size for later.

inRas = arcpy.Raster(in_raster)

x_cell = inRas.meanCellWidth
y_cell = inRas.meanCellHeight
ras_LLcorner = arcpy.Point(inRas.extent.XMin,inRas.extent.YMin)

# Create a Numpy Array from the input raster

npRas = arcpy.RasterToNumPyArray(inRas)

# Now figure out the PVCs required

# Check for commas
if pvc.find(",") == -1 and len(pvc) != 0:
    # NB any floating point values will be floored to the corresponding integer.
    pvc_list = [int(float(pvc))]
    if pvc_list[0] > 100 or pvc_list[0] < 1:
        arcpy.AddError("Invalid PVC entry")
        sys.exit()
elif pvc.find(",") > 0:
    # NB any floating point values will be floored to the corresponding integer.
    # split the string by commas
    pvc1 = pvc.split(",")
    # remove any empty strings
    pvc2 = filter(None,pvc1)
    # Map to float
    pvc3 = map(float,pvc2)
    # Finally Map to integers
    pvc_list = map(int,pvc3)
    if any(i > 100 for i in pvc_list) or any(j<1 for j in pvc_list):
        arcpy.AddError("Invalid PVC entry")
        sys.exit()
else:
    arcpy.AddError("Invalid PVC entry")
    sys.exit()

# Now threshold the numpy array

shp = npRas.shape

# Flatten the Array
flatArray = np.ravel(npRas)
del npRas

# Create cell volumes
volArray = (flatArray/np.sum(flatArray))*100.0
del flatArray

# Record an index for each cell, so we know their order after sorting
indexArray = np.dstack((np.arange(1,volArray.size+1),volArray))[0]
del volArray

# Sort the array by the densities column and reverse the order
sortArray = indexArray[indexArray[:,1].argsort()][::-1]
del indexArray

for p in pvc_list:
    cumSum = 0.0

    # This loop does the thresholding
    tempSortArray = copy.deepcopy(sortArray)
    
    for i in xrange(0,len(tempSortArray)):
        cumSum += tempSortArray[i][1]
        if cumSum <= p:
            tempSortArray[i][1] = 1.0
        else:
            tempSortArray[i][1] = 0.0
            
    # Resort the newly thresholded array
    reSortArray = tempSortArray[tempSortArray[:,0].argsort()]
    
    # Split the index column from the thresholded column
    splitArray = np.reshape(np.split(reSortArray,2,1)[1],shp)
    del reSortArray

    
    # For now, output the thresholded raster
    outRas = arcpy.NumPyArrayToRaster(splitArray,ras_LLcorner,x_cell,y_cell)
    #outRas.save(out_workspace + "\\rastpvc" + str(p))

    if polylines and not polygons:
        contour_shp = out_workspace + "\\PVC_Polyline_" + str(p) + ".shp"
        Contour(outRas,contour_shp,1,0.5)
    elif polygons and not polylines:
        contour_shp = out_workspace + "\\PVC_Polyline_" + str(p) + ".shp"
        Contour(outRas,contour_shp,1,0.5)
        arcpy.MakeFeatureLayer_management(contour_shp,"contour_shp_" + str(p))
        polygon_out = out_workspace + "\\PVC_Polygon_" + str(p) + ".shp"
        arcpy.FeatureToPolygon_management("contour_shp_" + str(p),polygon_out)
        if holes:
            removeholes(polygon_out,contour_shp)
        arcpy.Delete_management(contour_shp)
    elif polygons and polylines:
        contour_shp = out_workspace + "\\PVC_Polyline_" + str(p) + ".shp"
        Contour(outRas,contour_shp,1,0.5)
        arcpy.MakeFeatureLayer_management(contour_shp,"contour_shp_" + str(p))
        polygon_out = out_workspace + "\\PVC_Polygon_" + str(p) + ".shp"
        arcpy.FeatureToPolygon_management("contour_shp_" + str(p),polygon_out)
        #polygon_out1 = out_workspace + "\\PVC_Polygon_1_" + str(p) + ".shp"
        #arcpy.CopyFeatures_management(polygon_out,polygon_out1)
        if holes:
            removeholes(polygon_out,contour_shp)
    else:
        arcpy.AddError("Error Creating Outputs")
        sys.exit()

    del outRas
    del splitArray
        
    
